import json
import click
import cflow
import asyncio
import uvicorn


api = cflow.API()


async def run_execute(file: str, *, raw: bool = False, output: str = "") -> None:
    with open(file, "r") as f:
        data = cflow.WorkflowModel(**json.load(f))
    flow = data.get_workflow()
    results = await data.run(flow, return_api_response=not raw)
    if output:
        with open(output, "w") as f:
            json.dump(results, f, indent=2, ensure_ascii=False)


@click.group()
def main() -> None:
    pass


@main.command()
@click.option(
    "--app",
    default="cflow.cli:api.app",
    show_default=True,
    type=str,
    help="The target app of the server.",
)
@click.option(
    "-h",
    "--host",
    default="0.0.0.0",
    show_default=True,
    type=str,
    help="The target host of the server.",
)
@click.option(
    "-p",
    "--port",
    default=8123,
    show_default=True,
    type=int,
    help="The target port of the server.",
)
@click.option(
    "-r",
    "--reload",
    is_flag=True,
    default=False,
    show_default=True,
    type=bool,
    help="Whether to reload the server when the code changes.",
)
def serve(*, app: str, host: str, port: int, reload: bool) -> None:
    uvicorn.run(app, host=host, port=port, reload=reload)


@main.command()
@click.option(
    "-f",
    "--file",
    type=str,
    help="The workflow JSON file to be executed. It should follow the `WorkflowModel` schema.",
)
@click.option(
    "-o",
    "--output",
    default="",
    show_default=True,
    type=str,
    help="The output file path, please make sure that the outputs "
    "are JSON serializable if this option is specified.",
)
@click.option(
    "--raw",
    is_flag=True,
    default=False,
    show_default=True,
    type=bool,
    help="Whether to return the raw results instead of the API responses.",
)
def execute(*, file: str, output: str, raw: bool) -> None:
    asyncio.run(run_execute(file, raw=raw, output=output))


@main.command()
@click.option(
    "-f",
    "--file",
    type=str,
    help="The workflow JSON file to be executed. It should follow the `WorkflowModel` schema.",
)
@click.option(
    "-o",
    "--output",
    type=str,
    help="The output PNG file path.",
)
def render(*, file: str, output: str) -> None:
    with open(file, "r") as f:
        data = cflow.WorkflowModel(**json.load(f))
    rendered = cflow.render_workflow(data.get_workflow(), target=data.target)
    rendered.save(output)


__all__ = [
    "run_execute",
]



import json
import time
import asyncio

from abc import abstractmethod
from abc import ABCMeta
from PIL import Image
from cftool import console
from typing import Any
from typing import Set
from typing import Dict
from typing import List
from typing import Type
from typing import Union
from typing import TypeVar
from typing import Callable
from typing import Optional
from aiohttp import ClientSession
from pathlib import Path
from pydantic import BaseModel
from dataclasses import field
from dataclasses import dataclass
from cftool.web import get_err_msg
from cftool.web import download_raw_with_retry
from cftool.web import download_image_with_retry
from cftool.misc import offload
from cftool.misc import random_hash
from cftool.misc import register_core
from cftool.misc import shallow_copy_dict
from cftool.misc import ISerializableDataClass
from cftool.data_structures import Item
from cftool.data_structures import Bundle


TPath = Union[str, Path]
TTNode = TypeVar("TTNode", bound=Type["Node"])
nodes: Dict[str, Type["Node"]] = {}

UNDEFINED_PLACEHOLDER = "$undefined$"
EXCEPTION_MESSAGE_KEY = "$exception$"
ALL_LATENCIES_KEY = "$all_latencies$"
HTTP_SESSION_KEY = "$http_session$"

GATHER_NODE = "common.gather"


def extract_from(data: Any, hierarchy: str) -> Any:
    hierarchies = hierarchy.split(".")
    for h in hierarchies:
        if isinstance(data, list):
            try:
                ih = int(h)
            except:
                msg = f"current value is list, but '{h}' is not int"
                raise ValueError(msg)
            data = data[ih]
        elif isinstance(data, dict):
            data = data[h]
        else:
            raise ValueError(
                f"hierarchy '{h}' is required but current value type "
                f"is '{type(data)}' ({data})"
            )
    return data


@dataclass
class Injection:
    """
    A dataclass that represents an injection to the current node.

    Attributes
    ----------
    src_key : str
        The key of the dependent node.
    src_hierarchy : str | None
        The 'src_hierarchy' of the dependent node's results that the current node depends on.
        - `src_hierarchy` can be very complex:
          - use `int` as `list` index, and `str` as `dict` key.
          - use `.` to represent nested structure.
          - for example, you can use `a.0.b` to indicate `results["a"][0]["b"]`.
        - If `None`, all results of the dependent node will be used.
    dst_hierarchy : str
        The 'dst_hierarchy' of the current node's `data`.
        - `dst_hierarchy` can be very complex:
          - use `int` as `list` index, and `str` as `dict` key.
          - use `.` to represent nested structure.
          - for example, if you want to inject to `data["a"][0]["b"]`, you can use `a.0.b`
          as the `dst_hierarchy`.

    """

    src_key: str
    src_hierarchy: Optional[str]
    dst_hierarchy: str


@dataclass
class Schema:
    input_model: Optional[Type[BaseModel]] = None
    output_model: Optional[Type[BaseModel]] = None
    api_output_model: Optional[Type[BaseModel]] = None
    input_names: Optional[List[str]] = None
    output_names: Optional[List[str]] = None
    description: Optional[str] = None


@dataclass
class Node(ISerializableDataClass, metaclass=ABCMeta):
    """
    A Node class that represents a node in a workflow.

    This class is abstract and should be subclassed.

    Attributes
    ----------
    key : str, optional
        The key of the node, should be unique with respect to the workflow.
    data : Any, optional
        The data associated with the node.
    injections : List[Injection], optional
        A list of injections of the node.
    offload : bool, optional
        A flag indicating whether the node should be offloaded.
    lock_key : str, optional
        The lock key of the node.
    executing : bool, optional
        A runtime attribute indicating whether the node is currently executing.

    Methods
    -------
    async execute() -> Any
        Abstract method that should return the results.

    @classmethod
    get_schema() -> Optional[Schema]
        Optional method that returns the schema of the node.
        Implement this method can help us auto-generate UIs, APIs and documents.
    @classmethod
    async get_api_response(results: Dict[str, Any]) -> Any
        Optional method that returns the API response of the node from its 'raw' results.
        Implement this method to handle complex API responses (e.g. `PIL.Image`).
    async initialize(flow: Flow) -> None
        Optional method that will be called before the execution of the workflow.
        Implement this method to do heavy initializations (e.g. loading AI models).

    """

    key: Optional[str] = None
    data: Dict[str, Any] = field(default_factory=dict)
    injections: List[Injection] = field(default_factory=list)
    offload: bool = False
    lock_key: Optional[str] = None
    # runtime attribute, should not be touched and will not be serialized
    executing: bool = False
    shared_pool: Dict[str, Any] = field(default_factory=dict)

    # optional

    @classmethod
    def get_schema(cls) -> Optional[Schema]:
        return None

    @classmethod
    async def get_api_response(cls, results: Dict[str, Any]) -> Any:
        return results

    async def initialize(self, flow: "Flow") -> None:
        if HTTP_SESSION_KEY not in self.shared_pool:
            self.shared_pool[HTTP_SESSION_KEY] = ClientSession()

    async def cleanup(self) -> None:
        http_session = self.shared_pool.pop(HTTP_SESSION_KEY, None)
        if http_session is not None:
            await http_session.close()

    # abstract

    @abstractmethod
    async def execute(self) -> Any:
        pass

    # shortcuts

    @property
    def http_session(self) -> ClientSession:
        session = self.shared_pool.get(HTTP_SESSION_KEY)
        if session is None:
            raise ValueError(
                "`http_session` should be provided in the `shared_pool` "
                f"for `{self.__class__.__name__}`"
            )
        return session

    async def download_raw(self, url: str) -> bytes:
        return await download_raw_with_retry(self.http_session, url)

    async def download_image(self, url: str) -> Image.Image:
        return await download_image_with_retry(self.http_session, url)

    # internal

    @classmethod
    def d(cls) -> Dict[str, Type["Node"]]:
        return nodes

    @classmethod
    def register(cls, name: str) -> Callable[[TTNode], TTNode]:
        def before(cls_: Type) -> None:
            if name == "workflow":
                raise RuntimeError(
                    "`workflow` is a reserved name, please use another name "
                    f"when registering node '{cls_.__name__}'"
                )
            cls_.__identifier__ = name

        return register_core(
            name,
            cls.d(),
            allow_duplicate=False,
            before_register=before,
        )

    def copy(self) -> "Node":
        copied = self.__class__()
        copied.from_info(shallow_copy_dict(self.to_info()))
        copied.shared_pool = self.shared_pool
        return copied

    def to_item(self) -> Item["Node"]:
        if self.key is None:
            raise ValueError("node key cannot be None")
        return Item(self.key, self)

    def to_info(self) -> Dict[str, Any]:
        pool, self.shared_pool = self.shared_pool, {}
        info = super().to_info()
        info.pop("executing")
        info.pop("shared_pool")
        self.shared_pool = pool
        return info

    def from_info(self, info: Dict[str, Any]) -> None:
        super().from_info(info)
        self.injections = [Injection(**d) for d in self.injections]  # type: ignore
        if self.key is None:
            raise ValueError("node key cannot be None")
        if "." in self.key:
            raise ValueError("node key cannot contain '.'")

    def check_inputs(self) -> None:
        if not isinstance(self.data, dict):
            msg = f"input `data` ({self.data}) of node '{self.key}' should be a `dict`"
            raise ValueError(msg)
        schema = self.get_schema()
        if schema is None:
            return
        if schema.input_model is not None:
            try:
                narrowed = schema.input_model(**self.data)
                self.data = narrowed.model_dump()
            except Exception as err:
                msg = f"input data ({self.data}) does not match the schema model ({schema.input_model})"
                raise ValueError(msg) from err
        elif schema.input_names is not None:
            data_inputs = set(self.data.keys())
            schema_inputs = set(schema.input_names)
            if data_inputs != schema_inputs:
                msg = f"input data ({self.data}) does not match the schema names ({schema.input_names})"
                raise ValueError(msg)

    def check_injections(self) -> None:
        history: Dict[str, Injection] = {}
        for injection in self.injections:
            existing = history.get(injection.dst_hierarchy)
            if existing is not None:
                raise ValueError(
                    f"`dst_hierarchy` of current injection ({injection}) is duplicated "
                    f"with previous injection ({existing})"
                )
            history[injection.dst_hierarchy] = injection

    def fetch_injections(self, all_results: Dict[str, Any]) -> None:
        def inject_leaf_data(data: Any, hierarchies: List[str], value: Any) -> None:
            if len(hierarchies) == 0:
                return data
            h = hierarchies.pop(0)
            is_leaf = len(hierarchies) == 0
            if isinstance(data, list):
                try:
                    ih = int(h)
                except:
                    raise ValueError(f"current value is list, but '{h}' is not int")
                if len(data) <= ih:
                    replace_msg = "target value" if is_leaf else "an empty `dict`"
                    console.warn(
                        "current data is a list but its length is not enough, corresponding "
                        f"index ({h}) will be set to {replace_msg}, and other elements "
                        "will be set to `undefined`"
                    )
                    data.extend([UNDEFINED_PLACEHOLDER] * (ih - len(data) + 1))
                if is_leaf:
                    data[ih] = value
                else:
                    data[ih] = {}
                    inject_leaf_data(data[ih], hierarchies, value)
            elif isinstance(data, dict):
                if is_leaf:
                    data[h] = value
                else:
                    if h not in data:
                        console.warn(
                            "current data is a dict but it does not have the corresponding "
                            f"key ('{h}'), it will be set to an empty `dict`"
                        )
                        data[h] = {}
                    inject_leaf_data(data[h], hierarchies, value)
            else:
                raise ValueError(
                    f"hierarchy '{h}' is required but current value type "
                    f"is '{type(data)}' ({data})"
                )

        for injection in self.injections:
            src_key = injection.src_key
            src_out = all_results.get(src_key)
            if src_out is None:
                raise ValueError(f"cannot find cache for '{src_key}'")
            if injection.src_hierarchy is not None:
                src_out = extract_from(src_out, injection.src_hierarchy)
            dst_hierarchies = injection.dst_hierarchy.split(".")
            inject_leaf_data(self.data, dst_hierarchies, src_out)

    def check_undefined(self) -> None:
        def check(data: Any) -> None:
            if isinstance(data, list):
                for item in data:
                    check(item)
            elif isinstance(data, dict):
                for v in data.values():
                    check(v)
            elif data == UNDEFINED_PLACEHOLDER:
                raise ValueError(f"undefined value found in '{self.data}'")

        check(self.data)

    def check_results(self, results: Dict[str, Any]) -> Dict[str, Any]:
        if not isinstance(results, dict):
            msg = f"output results ({results}) of node '{self.key}' should be a `dict`"
            raise ValueError(msg)
        schema = self.get_schema()
        if schema is None:
            return results
        if schema.output_model is not None:
            try:
                narrowed = schema.output_model(**results)
                return narrowed.model_dump()
            except Exception as err:
                msg = f"output data ({results}) does not match the schema model ({schema.output_model})"
                raise ValueError(msg) from err
        if schema.output_names is not None:
            node_outputs = set(results.keys())
            schema_outputs = set(schema.output_names)
            if node_outputs != schema_outputs:
                msg = f"output data ({results}) does not match the schema names ({schema.output_names})"
                raise ValueError(msg)
        return results

    def check_api_results(self, results: Dict[str, Any]) -> Dict[str, Any]:
        schema = self.get_schema()
        if schema is None:
            return results
        if schema.api_output_model is not None:
            try:
                narrowed = schema.api_output_model(**results)
                return narrowed.model_dump()
            except Exception as err:
                msg = f"API response ({results}) does not match the schema model ({schema.api_output_model})"
                raise ValueError(msg) from err
        return results


class Flow(Bundle[Node]):
    """
    A Flow class that represents a workflow.

    Attributes
    ----------
    edges : Dict[str, List[Edge]]
        The dependencies of the workflow.
        - The key is the destination node key.
        - The value is a list of edges that indicates the dependencies
          of the destination node.
    latest_latencies : Dict[str, Dict[str, float]]
        The latest latencies of the workflow.

    Methods
    -------
    push(node: Node) -> Flow:
        Pushes a node into the workflow.
    gather(*targets: str) -> str:
        Gathers targets into a single node, and returns the key of the node.
    to_json() -> Dict[str, Any]:
        Converts the workflow to a JSON object.
    from_json(cls, data: Dict[str, Any]) -> Flow:
        Creates a workflow from a JSON object.
    dump(path: TPath) -> None:
        Dumps the workflow to a (JSON) file.
    load(cls, path: TPath) -> Flow:
        Loads a workflow from a (JSON) file.
    get_reachable(target: str) -> Set[str]:
        Gets the reachable nodes from a target.
    run(...) -> None:
        Runs a single node in the workflow.
    execute(...) -> Dict[str, Any]:
        Executes the entire workflow.

    """

    def __init__(self, *, no_mapping: bool = False) -> None:
        super().__init__(no_mapping=no_mapping)
        self.latest_latencies: Dict[str, Dict[str, float]] = {}

    def __str__(self) -> str:
        body = ",\n  ".join(str(item.data) for item in self)
        return f"""Flow([
  {body}
])"""

    __repr__ = __str__

    def push(self, node: Node) -> "Flow":
        if node.key is None:
            raise ValueError("node key cannot be None")
        super().push(node.to_item())
        return self

    def gather(self, *targets: str) -> str:
        gather_key = f"$gather_{random_hash()[:4]}"
        injections = [Injection(k, None, k) for k in targets]
        self.push(
            Node.make(
                GATHER_NODE,
                dict(key=gather_key, injections=injections),
            )
        )
        return gather_key

    def to_json(self) -> List[Dict[str, Any]]:
        return [item.data.to_pack().asdict() for item in self]

    @classmethod
    def from_json(cls, data: List[Dict[str, Any]]) -> "Flow":
        workflow = cls()
        for pack in data:
            workflow.push(Node.from_pack(pack))
        return workflow

    def copy(self) -> "Flow":
        return Flow.from_json(self.to_json())

    def dump(self, path: TPath) -> None:
        with open(path, "w") as f:
            json.dump(self.to_json(), f, indent=2)

    @classmethod
    def load(cls, path: TPath) -> "Flow":
        with open(path, "r") as f:
            return cls.from_json(json.load(f))

    def get_reachable(self, target: str) -> Set[str]:
        def dfs(key: str, is_target: bool) -> None:
            if not is_target and key == target:
                raise ValueError(f"cyclic dependency detected when dfs from '{target}'")
            if key in reachable:
                return
            reachable.add(key)
            item = self.get(key)
            if item is None:
                raise ValueError(
                    f"cannot find node '{key}', which is declared as a dependency, "
                    f"in the workflow ({self})"
                )
            node = item.data
            for injection in node.injections:
                dfs(injection.src_key, False)

        reachable: Set[str] = set()
        dfs(target, True)
        return reachable

    async def run(
        self,
        item: Item[Node],
        api_results: Dict[str, Any],
        all_results: Dict[str, Any],
        return_api_response: bool,
        verbose: bool,
        all_latencies: Dict[str, Dict[str, float]],
    ) -> None:
        if item.key in all_results:
            return
        start_t = time.time()
        while not all(i.src_key in all_results for i in item.data.injections):
            await asyncio.sleep(0)
        if item.data.lock_key is not None:
            while not all(
                not other.data.executing or other.data.lock_key != item.data.lock_key
                for other in self
            ):
                await asyncio.sleep(0)
        item.data.executing = True
        t0 = time.time()
        node: Node = item.data.copy()
        node.fetch_injections(all_results)
        node.check_undefined()
        node.check_inputs()
        t1 = time.time()
        if verbose:
            console.debug(f"executing node '{item.key}'")
        if not node.offload:
            results = await node.execute()
        else:
            results = await offload(node.execute())
        results = node.check_results(results)
        all_results[item.key] = results
        if return_api_response:
            results = await node.get_api_response(results)
            results = node.check_api_results(results)
            api_results[item.key] = results
        t2 = time.time()
        item.data.executing = False
        all_latencies[item.key] = dict(
            pending=t0 - start_t,
            inject=t1 - t0,
            execute=t2 - t1,
            latency=t2 - t0,
        )
        if verbose:
            console.debug(f"finished executing node '{item.key}'")

    async def execute(
        self,
        target: str,
        intermediate: Optional[List[str]] = None,
        *,
        return_api_response: bool = False,
        return_if_exception: bool = False,
        verbose: bool = False,
    ) -> Dict[str, Any]:
        """
        Executes the workflow ending at the `target` node.

        Parameters
        ----------
        target : str
            The key of the target node which the execution will end at.
        intermediate : List[str], optional
            A list of intermediate nodes that will be returned. Default is `None`.
            - Only useful when `return_api_response` is `True`.
            - If `None`, no intermediate nodes will be returned.
        return_if_exception : bool, optional
            If `True`, the function will return even if an exception occurs. Default is `False`.
        return_api_response : bool, optional
            If `True`, the function will:
            - Only return the results of the `target` node & the `intermediate` nodes.
            - Call `get_api_response` on the results to get the final API response.
        verbose : bool, optional
            If `True`, the function will print detailed logs. Default is `False`.

        Returns
        -------
        dict
            A dictionary containing the results of the execution. If an exception occurs
            during the execution, the dictionary will contain a key 'EXCEPTION_MESSAGE_KEY'
            with the error message as the value.

        """

        api_results: Dict[str, Any] = {}
        all_results: Dict[str, Any] = {}
        extra_results: Dict[str, Any] = {}
        all_latencies: Dict[str, Dict[str, float]] = {}
        if intermediate is None:
            intermediate = []
        try:
            workflow = self.copy()
            reachable = workflow.get_reachable(target)
            shared_pool: Dict[str, Any] = {}
            if target not in workflow:
                raise ValueError(f"cannot find target '{target}' in the workflow")
            reachable_nodes = [item.data for item in workflow if item.key in reachable]
            for node in reachable_nodes:
                node.check_injections()
                node.shared_pool = shared_pool
                if verbose:
                    console.debug(f"initializing node '{node.key}'")
                await node.initialize(workflow)
            await asyncio.gather(
                *(
                    workflow.run(
                        item,
                        api_results,
                        all_results,
                        return_api_response
                        and (item.key == target or item.key in intermediate),
                        verbose,
                        all_latencies,
                    )
                    for item in workflow
                    if item.key in reachable
                )
            )
            for node in reachable_nodes:
                if verbose:
                    console.debug(f"cleaning up node '{node.key}'")
                await node.cleanup()
            extra_results[EXCEPTION_MESSAGE_KEY] = None
        except Exception as err:
            if not return_if_exception:
                raise
            err_msg = get_err_msg(err)
            extra_results[EXCEPTION_MESSAGE_KEY] = err_msg
            if verbose:
                console.error(err_msg)
        self.latest_latencies = all_latencies
        extra_results[ALL_LATENCIES_KEY] = all_latencies
        final_results = api_results if return_api_response else all_results
        final_results.update(extra_results)
        return final_results


__all__ = [
    "HTTP_SESSION_KEY",
    "Injection",
    "Schema",
    "Node",
    "Flow",
    "BaseModel",
]



import json
import cflow
import unittest

from pathlib import Path
from fastapi.testclient import TestClient


file_folder = Path(__file__).parent
workflows_folder = file_folder.parent / "examples" / "workflows"


class TestServer(unittest.TestCase):
    def setUp(self):
        self.client = TestClient(cflow.cli.api.app)

    def test_server_status(self):
        response = self.client.get("/server_status")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json()["num_nodes"], len(cflow.use_all_t_nodes()))

    def test_workflows(self):
        for path in workflows_folder.glob("*.json"):
            with open(path, "r") as f:
                workflow = json.load(f)
            response = self.client.post("/workflow", json=workflow)
            self.assertEqual(response.status_code, 200)
            self.assertIsInstance(response.json(), dict)


if __name__ == "__main__":
    unittest.main()



# Computer Vision Nodes

import cv2

import numpy as np

from PIL import Image
from enum import Enum
from typing import Any, Dict
from typing import List
from typing import Tuple
from typing import Union
from typing import Optional
from pydantic import Field
from pydantic import BaseModel
from PIL.Image import Resampling as PILResampling
from PIL.ImageFilter import GaussianBlur
from cftool.cv import to_rgb
from cftool.cv import to_uint8
from cftool.cv import to_base64
from cftool.cv import ImageBox
from cftool.misc import safe_execute
from cftool.misc import shallow_copy_dict
from cftool.geometry import Matrix2D

from .common import TImage
from .common import ImageModel
from .common import IImageNode
from .common import IWithImageNode
from ..core import Node
from ..core import Schema


# common


class WHModel(BaseModel):
    w: int = Field(..., description="Width of the output image.")
    h: int = Field(..., description="Height of the output image")


class LTRBModel(BaseModel):
    lt_rb: Tuple[int, int, int, int] = Field(
        ...,
        description="The left-top and right-bottom points.",
    )


class ResizeMode(str, Enum):
    FILL = "fill"
    FIT = "fit"
    COVER = "cover"


class Resampling(str, Enum):
    NEAREST = "nearest"
    BOX = "box"
    BILINEAR = "bilinear"
    HAMMING = "hamming"
    BICUBIC = "bicubic"
    LANCZOS = "lanczos"


class ResamplingModel(BaseModel):
    resampling: Resampling = Field(Resampling.BILINEAR, description="The resampling.")


class BaseAffineModel(BaseModel):
    a: float = Field(..., description="`a` of the affine matrix")
    b: float = Field(..., description="`b` of the affine matrix")
    c: float = Field(..., description="`c` of the affine matrix")
    d: float = Field(..., description="`d` of the affine matrix")
    e: float = Field(..., description="`e` of the affine matrix")
    f: float = Field(..., description="`f` of the affine matrix")
    wh_limit: int = Field(
        16384,
        description="maximum width or height of the output image",
    )


def erode(array: np.ndarray, n_iter: int, kernel_size: int) -> np.ndarray:
    kernel = np.ones((kernel_size, kernel_size), np.uint8)
    return cv2.erode(array, kernel, iterations=n_iter)


def resize(image: Image.Image, w: int, h: int, resampling: Resampling) -> Image.Image:
    if resampling == Resampling.NEAREST:
        r = PILResampling.NEAREST
    elif resampling == Resampling.BOX:
        r = PILResampling.BOX
    elif resampling == Resampling.BILINEAR:
        r = PILResampling.BILINEAR
    elif resampling == Resampling.HAMMING:
        r = PILResampling.HAMMING
    elif resampling == Resampling.BICUBIC:
        r = PILResampling.BICUBIC
    else:
        r = PILResampling.LANCZOS
    return image.resize((w, h), r)


def affine(
    image: Image.Image,
    a: float,
    b: float,
    c: float,
    d: float,
    e: float,
    f: float,
    w: int,
    h: int,
    resampling: Resampling,
    wh_limit: int,
) -> np.ndarray:
    matrix2d = Matrix2D(a=a, b=b, c=c, d=d, e=e, f=f)
    properties = matrix2d.decompose()
    iw, ih = image.size
    nw = max(round(iw * abs(properties.w)), 1)
    nh = max(round(ih * abs(properties.h)), 1)
    if nw > wh_limit or nh > wh_limit:
        raise ValueError(f"image size ({nw}, {nh}) exceeds wh_limit ({wh_limit})")
    array = np.array(resize(image, nw, nh, resampling))
    properties.w = 1
    properties.h = 1 if properties.h > 0 else -1
    matrix2d = Matrix2D.from_properties(properties)
    return cv2.warpAffine(array, matrix2d.matrix, [w, h])


def paste(
    fg: Image.Image,
    bg: Image.Image,
    a: float,
    b: float,
    c: float,
    d: float,
    e: float,
    f: float,
    force_rgb: bool,
    resampling: Resampling,
    wh_limit: int,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    if fg.mode != "RGBA":
        fg = fg.convert("RGBA")
    w, h = bg.size
    affined = affine(fg, a, b, c, d, e, f, w, h, resampling, wh_limit)
    affined = affined.astype(np.float32) / 255.0
    rgb = affined[..., :3]
    mask = affined[..., -1:]
    if force_rgb:
        bg = to_rgb(bg)
    bg_array = np.array(bg).astype(np.float32) / 255.0
    fg_array = rgb if bg_array.shape[2] == 3 else affined
    pasted = fg_array * mask + bg_array * (1.0 - mask)
    return to_uint8(rgb), to_uint8(mask[..., 0]), to_uint8(pasted)


def get_contrast_bg(rgba_image: Image.Image) -> int:
    rgba = np.array(rgba_image)
    rgb = rgba[..., :3]
    alpha = rgba[..., -1]
    target_mask = alpha >= 10
    hls = cv2.cvtColor(rgb, cv2.COLOR_RGB2HLS)
    lightness = hls[..., 1].astype(np.float32) / 255.0
    target_lightness = lightness[target_mask]
    mean = target_lightness.mean().item()
    std = target_lightness.std().item()
    if 0.45 <= mean <= 0.55 and std >= 0.25:
        return 127
    if mean <= 0.2 or 0.8 <= mean:
        return 127
    return 0 if mean >= 0.5 else 255


# nodes


class BinaryOpInput(ImageModel):
    url2: TImage = Field(..., description="The second image to process.")
    op: str = Field(..., description="The (numpy) operation.")


@Node.register("cv.binary_op")
class BinaryOpNode(IImageNode):
    @classmethod
    def get_schema(cls) -> Schema:
        schema = super().get_schema()
        schema.input_model = BinaryOpInput
        schema.description = "Perform binary (numpy) operations on two images."
        return schema

    async def execute(self) -> Dict[str, Image.Image]:
        op_name = self.data["op"]
        op = getattr(np, op_name, None)
        if op is None:
            raise ValueError(f"`{op_name}` is not a valid numpy operation")
        array = np.array(await self.get_image_from("url"))
        array2 = np.array(await self.get_image_from("url2"))
        image = Image.fromarray(op(array, array2))
        return {"image": image}


class BlurInput(ImageModel):
    radius: int = Field(2, description="Size of the blurring kernel.")


@Node.register("cv.blur")
class BlurNode(IImageNode):
    @classmethod
    def get_schema(cls) -> Schema:
        schema = super().get_schema()
        schema.input_model = BlurInput
        schema.description = "Blur an image with a Gaussian filter."
        return schema

    async def execute(self) -> Dict[str, Image.Image]:
        image = await self.get_image_from("url")
        image = image.filter(GaussianBlur(radius=self.data["radius"]))
        return {"image": image}


@Node.register("cv.inverse")
class InverseNode(IImageNode):
    @classmethod
    def get_schema(cls) -> Schema:
        schema = super().get_schema()
        schema.description = "Inverse an image."
        return schema

    async def execute(self) -> Dict[str, Image.Image]:
        image = await self.get_image_from("url")
        image = Image.fromarray(255 - np.array(image))
        return {"image": image}


@Node.register("cv.grayscale")
class GrayscaleNode(IImageNode):
    @classmethod
    def get_schema(cls) -> Schema:
        schema = super().get_schema()
        schema.description = "Convert an image to grayscale."
        return schema

    async def execute(self) -> Dict[str, Image.Image]:
        image = await self.get_image_from("url")
        image = image.convert("L")
        return {"image": image}


class ErodeAlphaInput(ImageModel):
    n_iter: int = Field(1, description="Number of iterations.")
    kernel_size: int = Field(3, description="Size of the kernel.")
    threshold: int = Field(0, description="Threshold of the alpha channel.")
    padding: int = Field(8, description="Padding for the image.")


@Node.register("cv.erode_alpha")
class ErodeAlphaNode(IImageNode):
    @classmethod
    def get_schema(cls) -> Schema:
        schema = super().get_schema()
        schema.input_model = ErodeAlphaInput
        schema.description = "Erode an image."
        return schema

    async def execute(self) -> Dict[str, Image.Image]:
        n_iter = self.data["n_iter"]
        kernel_size = self.data["kernel_size"]
        threshold = self.data["threshold"]
        padding = self.data["padding"]
        image = await self.get_image_from("url")
        w, h = image.size
        array = np.array(image.convert("RGBA"))
        alpha = array[..., -1]
        padded = np.pad(alpha, (padding, padding), constant_values=0)
        binarized = cv2.threshold(padded, threshold, 255, cv2.THRESH_BINARY)[1]
        eroded = erode(binarized, n_iter, kernel_size)
        shrinked = eroded[padding:-padding, padding:-padding]
        merged_alpha = np.minimum(alpha, shrinked)
        array[..., -1] = merged_alpha
        rgb = array[..., :3].reshape([-1, 3])
        rgb[(merged_alpha == 0).ravel()] = 0
        array[..., :3] = rgb.reshape([h, w, 3])
        image = Image.fromarray(array)
        return {"image": image}


class ResizeInput(ResamplingModel, WHModel, ImageModel):
    mode: ResizeMode = Field(ResizeMode.FILL, description="The Resize mode.")


@Node.register("cv.resize")
class ResizeNode(IImageNode):
    @classmethod
    def get_schema(cls) -> Schema:
        schema = super().get_schema()
        schema.input_model = ResizeInput
        schema.description = "Resize an image."
        return schema

    async def execute(self) -> Dict[str, Image.Image]:
        w = self.data["w"]
        h = self.data["h"]
        mode = self.data["mode"]
        resampling = self.data["resampling"]
        image = await self.get_image_from("url")
        img_w, img_h = image.size
        if mode != ResizeMode.FILL:
            w_ratio = w / img_w
            h_ratio = h / img_h
            if mode == ResizeMode.FIT:
                ratio = min(w_ratio, h_ratio)
            else:
                ratio = max(w_ratio, h_ratio)
            w = round(img_w * ratio)
            h = round(img_h * ratio)
        image = resize(image, w, h, resampling)
        return {"image": image}


class AffineInput(ResamplingModel, WHModel, BaseAffineModel, ImageModel):
    force_rgba: bool = Field(
        False,
        description="Whether to force the output to be RGBA.",
    )


@Node.register("cv.affine")
class AffineNode(IImageNode):
    @classmethod
    def get_schema(cls) -> Schema:
        schema = super().get_schema()
        schema.input_model = AffineInput
        schema.description = "Affine an image."
        return schema

    async def execute(self) -> Dict[str, Image.Image]:
        image = await self.get_image_from("url")
        if self.data["force_rgba"] and image.mode != "RGBA":
            image = image.convert("RGBA")
        affine_kw = shallow_copy_dict(self.data)
        affine_kw["image"] = image
        array = safe_execute(affine, affine_kw)
        image = Image.fromarray(array)
        return {"image": image}


class GetMaskInput(ImageModel):
    get_inverse: bool = Field(False, description="Whether get the inverse mask.")
    binarize_threshold: Optional[int] = Field(
        None,
        ge=0,
        le=255,
        description="If not `None`, will binarize the mask with this value.",
    )


@Node.register("cv.get_mask")
class GetMaskNode(IImageNode):
    @classmethod
    def get_schema(cls) -> Schema:
        schema = super().get_schema()
        schema.input_model = GetMaskInput
        schema.description = "Get the mask of an image."
        return schema

    async def execute(self) -> Dict[str, Image.Image]:
        image = await self.get_image_from("url")
        if image.mode == "RGBA":
            mask = np.array(image)[..., -1]
        else:
            mask = np.array(image.convert("L"))
        if self.data["get_inverse"]:
            mask = 255 - mask
        binarize_threshold = self.data["binarize_threshold"]
        if binarize_threshold is not None:
            mask = np.where(mask > binarize_threshold, 255, 0)
            mask = mask.astype(np.uint8)
        mask_image = Image.fromarray(mask)
        return {"image": mask_image}


class FillBGInput(ImageModel):
    bg: Optional[Union[int, Tuple[int, int, int]]] = Field(
        None,
        description="""Target background color.
> If not specified, `get_contrast_bg` will be used to calculate the `bg`.
""",
    )


@Node.register("cv.fill_bg")
class FillBGNode(IImageNode):
    @classmethod
    def get_schema(cls) -> Schema:
        schema = super().get_schema()
        schema.input_model = FillBGInput
        schema.description = "Fill the background of an image."
        return schema

    async def execute(self) -> Dict[str, Image.Image]:
        image = await self.get_image_from("url")
        bg = self.data["bg"]
        if bg is None:
            bg = get_contrast_bg(image)
        if isinstance(bg, int):
            bg = bg, bg, bg
        image = Image.fromarray(np.array(image))
        image = to_rgb(image, bg)
        return {"image": image}


class GetSizeOutput(BaseModel):
    size: Tuple[int, int] = Field(..., description="The size of the image.")


@Node.register("cv.get_size")
class GetSizeNode(IWithImageNode):
    @classmethod
    def get_schema(cls) -> Schema:
        return Schema(
            ImageModel,
            GetSizeOutput,
            description="Get the size of an image.",
        )

    async def execute(self) -> Dict[str, Tuple[int, int]]:
        image = await self.get_image_from("url")
        return {"size": image.size}


class ModifyBoxInput(LTRBModel):
    w: Optional[int] = Field(None, description="The width of the image.")
    h: Optional[int] = Field(None, description="The height of the image.")
    padding: int = Field(0, description="The padding size.")
    to_square: bool = Field(False, description="Turn the box into a square box.")


@Node.register("cv.modify_box")
class ModifyBoxNode(Node):
    @classmethod
    def get_schema(cls) -> Schema:
        return Schema(ModifyBoxInput, LTRBModel, description="Modify the box.")

    async def execute(self) -> Dict[str, Tuple[int, int, int, int]]:
        box = ImageBox(*self.data["lt_rb"])
        box = box.pad(self.data["padding"], w=self.data["w"], h=self.data["h"])
        if self.data["to_square"]:
            box = box.to_square()
        return {"lt_rb": box.tuple}


class GenerateMasksInput(BaseModel):
    w: int = Field(..., description="The width of the canvas.")
    h: int = Field(..., description="The height of the canvas.")
    lt_rb_list: List[Tuple[int, int, int, int]] = Field(..., description="The boxes.")
    merge: bool = Field(False, description="Whether merge the masks.")


class GenerateMaskAPIOutput(BaseModel):
    masks: List[str] = Field(..., description="The base64 encoded masks.")


@Node.register("cv.generate_masks")
class GenerateMasksNode(Node):
    @classmethod
    def get_schema(cls) -> Schema:
        return Schema(
            GenerateMasksInput,
            api_output_model=GenerateMaskAPIOutput,
            output_names=["masks"],
            description="Generate mask images from boxes.",
        )

    @classmethod
    async def get_api_response(cls, results: Dict[str, Any]) -> Dict[str, List[str]]:
        masks = list(map(to_base64, results["masks"]))
        return {"masks": masks}

    async def execute(self) -> Dict[str, List[Image.Image]]:
        merge = self.data["merge"]
        canvas = np.zeros((self.data["h"], self.data["w"]), dtype=np.uint8)
        results = []
        for l, t, r, b in self.data["lt_rb_list"]:
            i_canvas = canvas if merge else canvas.copy()
            i_canvas[t:b, l:r] = 255
            if not merge:
                results.append(Image.fromarray(i_canvas))
        if merge:
            results.append(Image.fromarray(canvas))
        return {"masks": results}


class CropImageInput(LTRBModel, ImageModel):
    pass


@Node.register("cv.crop_image")
class CropImageNode(IImageNode):
    @classmethod
    def get_schema(cls) -> Schema:
        schema = super().get_schema()
        schema.input_model = CropImageInput
        schema.description = "Crop an image."
        return schema

    async def execute(self) -> Dict[str, Image.Image]:
        image = await self.get_image_from("url")
        image = image.crop(self.data["lt_rb"])
        return {"image": image}


class PasteModel(BaseModel):
    bg_url: TImage = Field(..., description="The background image.")
    force_rgb: bool = Field(False, description="Whether to force the output to be RGB.")


class PasteInput(ResamplingModel, BaseAffineModel, PasteModel, ImageModel):
    pass


class PasteAPIOutput(BaseModel):
    rgb: str = Field(
        ...,
        description="The base64 encoded RGB image.\n"
        "> This is the affined `fg` which does not contain the background.",
    )
    mask: str = Field(
        ...,
        description="The base64 encoded mask image.\n"
        "> This is the affined `fg`'s alpha channel.",
    )
    pasted: str = Field(..., description="The base64 encoded pasted image.")


@Node.register("cv.paste")
class PasteNode(IWithImageNode):
    @classmethod
    def get_schema(cls) -> Schema:
        return Schema(
            PasteInput,
            api_output_model=PasteAPIOutput,
            output_names=["rgb", "mask", "pasted"],
            description="Paste an image onto another image.",
        )

    @classmethod
    async def get_api_response(cls, results: Dict[str, Any]) -> Dict[str, str]:
        return {k: to_base64(v) for k, v in results.items()}

    async def execute(self) -> Dict[str, Image.Image]:
        paste_kw = shallow_copy_dict(self.data)
        paste_kw["fg"] = await self.get_image_from("url")
        paste_kw["bg"] = await self.get_image_from("bg_url")
        rgb, mask, pasted = safe_execute(paste, paste_kw)
        rgb = Image.fromarray(rgb)
        mask = Image.fromarray(mask)
        pasted = Image.fromarray(pasted)
        return {"rgb": rgb, "mask": mask, "pasted": pasted}


__all__ = [
    "BinaryOpNode",
    "BlurNode",
    "InverseNode",
    "GrayscaleNode",
    "ErodeAlphaNode",
    "ResizeNode",
    "AffineNode",
    "GetMaskNode",
    "FillBGNode",
    "GetSizeNode",
    "ModifyBoxNode",
    "GenerateMasksNode",
    "CropImageNode",
]



# Common Nodes

import asyncio

from PIL import Image
from cftool import console
from typing import Any
from typing import Dict
from typing import List
from typing import Union
from typing import Optional
from pydantic import Field
from pydantic import BaseModel
from dataclasses import dataclass
from pydantic_core import core_schema
from cftool.cv import to_base64
from cftool.misc import shallow_copy_dict
from cftool.console import log

from ..core import GATHER_NODE
from ..core import extract_from
from ..core import Node
from ..core import Flow
from ..core import Schema
from ..core import Injection


# functional nodes


class LoopBackInjection(BaseModel):
    """
    A dataclass that represents a loop back injection to the current node.

    > This is the same as `Injection`, except the `src_key` will always be the
    key of the previous node in the loop.
    """

    src_hierarchy: Optional[str] = Field(
        ...,
        description="""The 'src_hierarchy' of the dependent node's results that the current node depends on.
- `src_hierarchy` can be very complex:
  - use `int` as `list` index, and `str` as `dict` key.
  - use `.` to represent nested structure.
  - for example, you can use `a.0.b` to indicate `results["a"][0]["b"]`.
- If `None`, all results of the dependent node will be used.""",
    )
    dst_hierarchy: str = Field(
        ...,
        description="""The 'dst_hierarchy' of the current node's `data`.
- `dst_hierarchy` can be very complex:
  - use `int` as `list` index, and `str` as `dict` key.
  - use `.` to represent nested structure.
  - for example, if you want to inject to `data["a"][0]["b"]`, you can use `a.0.b`
  as the `dst_hierarchy`.""",
    )


class LoopInput(BaseModel):
    base_node: str = Field(..., description="The node to be looped.")
    base_data: Dict[str, Any] = Field(default_factory=dict, description="Base data.")
    loop_values: Dict[str, List[Any]] = Field(
        ...,
        description="""The values to be looped.
> - The keys should be the 'target hierarchy' of the `data`
> - The values should be a list of values to be looped & injectedinto the 'target hierarchy'.
> - All values should have the same length.

For example, if you want to loop `data["a"]` with values `[1, 2]`, and loop `data["b"][0]["c"]` with values `[3, 4]`, you can use:
```python
{
    "a": [1, 2],
    "b.0.c": [3, 4],
}
```
""",
    )
    loop_back_injections: Optional[List[LoopBackInjection]] = Field(
        None,
        description="The loop back injections.\n"
        "> - If this is set, the results from the previous step in the loop will be "
        "injected into the current node's `data`.\n"
        "> - If `None`, no injection will be performed, and all nodes will be "
        "executed in parallel.",
    )
    extract_hierarchy: Optional[str] = Field(
        None,
        description="The hierarchy of the results to be extracted.\n"
        "> - If `None`, all results will be preserved.",
    )
    verbose: bool = Field(False, description="Whether to print debug logs.")


class LoopOutput(BaseModel):
    results: List[Any] = Field(..., description="The results of the loop.")


@Node.register("common.loop")
class LoopNode(Node):
    @classmethod
    def get_schema(cls) -> Schema:
        return Schema(
            LoopInput,
            LoopOutput,
            description="A node that represents a loop of another node.",
        )

    async def execute(self) -> Dict[str, List[Dict[str, Any]]]:
        t_node = Node.get(self.data["base_node"])
        if t_node is None:
            raise ValueError(f"node `{self.data['base_node']}` is not defined")
        base_data = self.data["base_data"]
        loop_values = self.data["loop_values"]
        loop_back_injections = self.data["loop_back_injections"]
        loop_keys = list(loop_values)
        lengths = [len(loop_values[k]) for k in loop_keys]
        if len(set(lengths)) != 1:
            raise ValueError(
                "all loop values should have the same length, "
                f"but lengths are {lengths}"
            )
        n = lengths[0]
        flow = Flow()
        for i in range(n):
            i_data = shallow_copy_dict(base_data)
            for k in loop_keys:
                i_data[k] = loop_values[k][i]
            if loop_back_injections is None:
                i_injections = []
            else:
                i_injections = list(map(shallow_copy_dict, loop_back_injections))
                i_injections = [Injection(**d) for d in i_injections]
            flow.push(t_node(str(i), i_data, i_injections))
        target = flow.gather(*map(str, range(n)))
        results = await flow.execute(target, verbose=self.data["verbose"])
        extracted = [results[str(i)] for i in range(n)]
        extract_hierarchy = self.data["extract_hierarchy"]
        if extract_hierarchy is not None:
            extracted = [extract_from(rs, extract_hierarchy) for rs in extracted]
        return {"results": extracted}


@Node.register(GATHER_NODE)
class GatherNode(Node):
    flow: Optional[Flow] = None

    @classmethod
    async def get_api_response(cls, results: Dict[str, Any]) -> Dict[str, Any]:
        if cls.flow is None:
            console.warn(
                "`flow` is not provided for `GatherNode`, raw results will be returned "
                "and `get_api_response` might not work as expected"
            )
            return results
        keys = list(results)
        node_names = [cls.flow.get(k).data.__identifier__ for k in keys]
        t_nodes = list(map(Node.get, node_names))
        tasks = [t.get_api_response(results[k]) for k, t in zip(keys, t_nodes)]
        converted = await asyncio.gather(*tasks)
        return {k: v for k, v in zip(keys, converted)}

    async def initialize(self, flow: Flow) -> None:
        await super().initialize(flow)
        self.flow = flow

    async def execute(self) -> Dict[str, Any]:
        return self.data

    def from_info(self, info: Dict[str, Any]) -> None:
        super().from_info(info)
        for injection in self.injections:
            if injection.src_hierarchy is not None:
                raise ValueError(
                    "`GatherNode` should always use `src_hierarchy=None` "
                    f"for injections, but `{injection}` is found"
                )
            if injection.src_key != injection.dst_hierarchy:
                raise ValueError(
                    "`GatherNode` should always use `src_key=dst_hierarchy` "
                    f"for injections, but `{injection}` is found"
                )


# common fields / data models


class ImageField(Image.Image):
    @classmethod
    def __get_pydantic_core_schema__(cls, *args: Any) -> core_schema.CoreSchema:
        return core_schema.with_info_plain_validator_function(cls.validate)

    @classmethod
    def validate(cls, v: Any, info: core_schema.ValidationInfo) -> Image.Image:
        if isinstance(v, Image.Image):
            return v
        raise ValueError("Value must be a PIL Image")


TImage = Union[str, ImageField]


class ImageModel(BaseModel):
    url: TImage = Field(..., description="The url / PIL.Image instance of the image.")


class ImageAPIOuput(BaseModel):
    image: str = Field(..., description="The base64 encoded image.")


class EmptyOutput(BaseModel):
    pass


# common node interfaces


class IWithImageNode(Node):
    """
    node interface which has image(s) as input. This is helpful for crafting image processing nodes.

    Notes
    -----
    - This interface provides `get_image_from` to get the image from the given field.
      - If the field is a url, it will be downloaded.
      - The field can also be `PIL.Image` directly, since it might be injected by
        other nodes.

    """

    async def fetch_image(self, tag: str, image: TImage) -> Image.Image:
        if isinstance(image, str):
            image = await self.download_image(image)
        elif not isinstance(image, Image.Image):
            raise ValueError(f"`{tag}` should be a `PIL.Image` or a url")
        return image

    async def get_image_from(self, field: str) -> Image.Image:
        image = self.data[field]
        return await self.fetch_image(field, image)


class IImageNode(IWithImageNode):
    """
    Image node interface. This is helpful for crafting image processing nodes.

    Notes
    -----
    - This interface assumes the output to be like `{"image": PIL.Image}`.

    """

    @classmethod
    def get_schema(cls) -> Schema:
        return Schema(
            ImageModel,
            api_output_model=ImageAPIOuput,
            output_names=["image"],
        )

    @classmethod
    async def get_api_response(cls, results: Dict[str, Image.Image]) -> Dict[str, str]:
        return {"image": to_base64(results["image"])}


@dataclass
class ICUDANode(Node):
    """
    CUDA node interface. This is helpful when creating nodes for modern AI models.

    Notes
    -----
    - CUDA executions should be 'offloaded' to avoid blocking other async executions.
    - CUDA executions should be 'locked' to avoid CUDA issues.
    """

    offload: bool = True
    lock_key: str = "$cuda$"


# common nodes


class ParametersModel(BaseModel):
    params: Dict[str, Any] = Field(default_factory=dict, description="The parameters.")


@Node.register("common.parameters")
class ParametersNode(Node):
    @classmethod
    def get_schema(cls) -> Schema:
        return Schema(
            ParametersModel,
            ParametersModel,
            description="Setup parameters.\n"
            "> - This is often used in a pre-defined workflow JSON to decide "
            "which parameters to be exposed to the user.\n"
            "> - See [examples](https://github.com/carefree0910/carefree-workflow/tree/main/examples/workflows) for reference.",
        )

    async def execute(self) -> Dict[str, Any]:
        return self.data


class EchoModel(BaseModel):
    messages: Union[str, List[str]]


@Node.register("common.echo")
class EchoNode(Node):
    @classmethod
    def get_schema(cls) -> Schema:
        return Schema(EchoModel, EchoModel, description="Echo the given message(s).")

    async def execute(self) -> Dict[str, Union[str, List[str]]]:
        messages = self.data["messages"]
        if isinstance(messages, str):
            messages = [messages]
        for message in messages:
            log(message)
        return self.data


@dataclass
@Node.register("common.download_image")
class DownloadImageNode(IImageNode):
    offload: bool = True

    @classmethod
    def get_schema(cls) -> Schema:
        schema = super().get_schema()
        schema.description = "Download an image from the given url."
        return schema

    async def execute(self) -> Dict[str, Image.Image]:
        image = await self.get_image_from("url")
        return {"image": image}


class SaveImageInput(ImageModel):
    path: str = Field("debug.png", description="The path to save the image.")


@Node.register("debug.save_image")
class SaveImageNode(IWithImageNode):
    @classmethod
    def get_schema(cls) -> Schema:
        return Schema(
            SaveImageInput,
            output_model=EmptyOutput,
            description="Save an image from the given url to disk, mainly for debugging.",
        )

    async def execute(self) -> dict:
        image = await self.get_image_from("url")
        image.save(self.data["path"])
        return {}


class SaveImagesInput(BaseModel):
    urls: List[TImage] = Field(..., description="The urls of the images.")
    prefix: str = Field("debug", description="The prefix to save the images.")


@Node.register("debug.save_images")
class SaveImagesNode(IWithImageNode):
    @classmethod
    def get_schema(cls) -> Schema:
        return Schema(
            SaveImagesInput,
            output_model=EmptyOutput,
            description="Save images from the given urls to disk, mainly for debugging.",
        )

    async def execute(self) -> dict:
        tasks = [self.fetch_image(str(v), v) for v in self.data["urls"]]
        images = await asyncio.gather(*tasks)
        prefix = self.data["prefix"]
        for i, image in enumerate(images):
            image.save(f"{prefix}_{i}.png")
        return {}


__all__ = [
    "LoopNode",
    "GatherNode",
    "TImage",
    "ImageModel",
    "ImageAPIOuput",
    "IImageNode",
    "ICUDANode",
    "EchoNode",
    "DownloadImageNode",
    "SaveImageNode",
    "SaveImagesNode",
]



import unittest
import subprocess

from pathlib import Path


file_folder = Path(__file__).parent
examples_folder = file_folder.parent / "examples"


class TestExample(unittest.TestCase):
    def test_code_snippets(self):
        for path in examples_folder.glob("*.py"):
            subprocess.run(["python", str(path)], check=True)

    def test_workflows(self):
        for path in (examples_folder / "workflows").rglob("*.json"):
            subprocess.run(
                [
                    "cflow",
                    "execute",
                    "-f",
                    str(path),
                    "-o",
                    f"{path.stem}_results.json",
                ],
                check=True,
            )
            subprocess.run(
                ["cflow", "render", "-f", str(path), "-o", f"{path.stem}.png"],
                check=True,
            )


if __name__ == "__main__":
    unittest.main()



from typing import Any
from typing import Dict
from typing import List
from typing import Type
from typing import Optional
from fastapi import FastAPI
from pydantic import create_model
from pydantic import Field
from pydantic import BaseModel
from cftool.web import raise_err
from cftool.web import get_responses
from cftool.misc import random_hash

from .core import Node
from .core import Flow
from .nodes.common import LoopBackInjection


def parse_endpoint(t_node: Type[Node]) -> str:
    split = t_node.__identifier__.split(".")
    return f"/{'/'.join(split)}"


def parse_input_model(t_node: Type[Node]) -> Optional[Type[BaseModel]]:
    schema = t_node.get_schema()
    if schema is None:
        return None
    if schema.input_model is not None:
        return schema.input_model
    if schema.input_names is not None:
        return create_model(  # type: ignore
            f"{t_node.__name__}Input",
            **{name: (Any, ...) for name in schema.input_names},
        )
    return None


def parse_output_model(t_node: Type[Node]) -> Optional[Type[BaseModel]]:
    schema = t_node.get_schema()
    if schema is None:
        return None
    if schema.api_output_model is not None:
        return schema.api_output_model
    if schema.output_model is not None:
        return schema.output_model
    if schema.output_names is not None:
        return create_model(  # type: ignore
            f"{t_node.__name__}Output",
            **{name: (Any, ...) for name in schema.output_names},
        )
    return None


def parse_description(t_node: Type[Node]) -> Optional[str]:
    schema = t_node.get_schema()
    if schema is None:
        return None
    return schema.description


def use_all_t_nodes() -> List[Type[Node]]:
    return list(t_node for t_node in Node.d().values() if issubclass(t_node, Node))


def register_api(app: FastAPI, t_node: Type[Node]) -> None:
    endpoint = parse_endpoint(t_node)
    input_model = parse_input_model(t_node)
    output_model = parse_output_model(t_node)
    description = parse_description(t_node)
    if input_model is None or output_model is None:
        return None
    names = t_node.__identifier__.split(".")
    names[0] = f"[{names[0]}]"
    name = "_".join(names)

    @app.post(
        endpoint,
        name=name,
        responses=get_responses(output_model),
        description=description,
    )
    async def _(data: input_model) -> output_model:  # type: ignore
        try:
            node = t_node(random_hash(), data.model_dump())  # type: ignore
            results = await node.execute()
            results = await t_node.get_api_response(results)
            return output_model(**results)
        except Exception as err:
            raise_err(err)


def register_nodes_api(app: FastAPI) -> None:
    for t_node in use_all_t_nodes():
        register_api(app, t_node)


class SrcKey(BaseModel):
    src_key: str = Field(..., description="The key of the dependent node.")


class InjectionModel(LoopBackInjection, SrcKey):
    pass


class NodeModel(BaseModel):
    key: str = Field(
        ...,
        description="The key of the node, should be unique with respect to the workflow.",
    )
    name: str = Field(
        ...,
        description="The name of the node, should be the one when registered.",
    )
    data: Dict[str, Any] = Field(
        default_factory=dict,
        description="The data associated with the node.",
    )
    injections: List[InjectionModel] = Field(
        default_factory=list,
        description="A list of injections of the node.",
    )
    offload: bool = Field(
        False,
        description="A flag indicating whether the node should be offloaded.",
    )
    lock_key: Optional[str] = Field(None, description="The lock key of the node.")


class WorkflowModel(BaseModel):
    target: str = Field(..., description="The target output node of the workflow.")
    intermediate: Optional[List[str]] = Field(
        None,
        description="The intermediate nodes that you want to get outputs from.",
    )
    nodes: List[NodeModel] = Field(..., description="A list of nodes in the workflow.")
    return_if_exception: bool = Field(
        False,
        description="Whether to return partial results if exception occurs.",
    )
    verbose: bool = Field(False, description="Whether to print debug logs.")

    def get_workflow(self) -> Flow:
        workflow_json = []
        for node in self.model_dump()["nodes"]:
            node_json = dict(type=node.pop("name"), info=node)
            workflow_json.append(node_json)
        return Flow.from_json(workflow_json)

    async def run(
        self,
        flow: Flow,
        *,
        return_api_response: bool = False,
    ) -> Dict[str, Any]:
        return await flow.execute(
            self.target,
            self.intermediate,
            return_api_response=return_api_response,
            return_if_exception=self.return_if_exception,
            verbose=self.verbose,
        )


def register_workflow_api(app: FastAPI) -> None:
    @app.post("/workflow")
    async def workflow(data: WorkflowModel) -> Dict[str, Any]:
        try:
            flow = data.get_workflow()
            return await data.run(flow, return_api_response=True)
        except Exception as err:
            raise_err(err)
            return {}


class ServerStatus(BaseModel):
    num_nodes: int = Field(
        ...,
        description="The number of registered nodes in the environment.\n"
        "> - Notice that this may be different from the number of nodes "
        "which are exposed as API, because some nodes may not have "
        "`get_schema` method implemented.\n"
        "> - However, all nodes can be used in the `workflow` API, no matter "
        "whether they have `get_schema` method implemented or not.",
    )


def register_server_api(app: FastAPI) -> None:
    @app.get("/server_status", responses=get_responses(ServerStatus))
    async def server_status() -> ServerStatus:
        return ServerStatus(num_nodes=len(use_all_t_nodes()))


class API:
    def __init__(self) -> None:
        self.app = FastAPI()
        register_server_api(self.app)
        register_nodes_api(self.app)
        register_workflow_api(self.app)


__all__ = [
    "parse_endpoint",
    "parse_input_model",
    "parse_output_model",
    "use_all_t_nodes",
    "register_api",
    "register_nodes_api",
    "register_workflow_api",
    "WorkflowModel",
    "API",
]



import io

import networkx as nx
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches

from PIL import Image
from typing import Set
from typing import Dict
from typing import List
from typing import Tuple
from typing import Optional
from typing import NamedTuple
from cftool.data_structures import Item

from .core import Node
from .core import Flow


class ToposortResult(NamedTuple):
    in_edges: Dict[str, Set[str]]
    hierarchy: List[List[Item[Node]]]
    edge_labels: Dict[Tuple[str, str], str]
    reachable: Set[str]


def toposort(workflow: Flow) -> ToposortResult:
    in_edges: Dict[str, Set[str]] = {item.key: set() for item in workflow}
    out_degrees = {item.key: 0 for item in workflow}
    edge_labels: Dict[Tuple[str, str], str] = {}
    for item in workflow:
        visited = set()
        dst_key = item.key
        for injection in item.data.injections:
            in_edges[injection.src_key].add(dst_key)
            if injection.src_key not in visited:
                visited.add(injection.src_key)
                out_degrees[dst_key] += 1
            label_key = (dst_key, injection.src_key)
            edge_label = injection.dst_hierarchy
            existing_label = edge_labels.get(label_key)
            if existing_label is None:
                edge_labels[label_key] = edge_label
            else:
                edge_labels[label_key] = f"{existing_label}, {edge_label}"

    ready = [k for k, v in out_degrees.items() if v == 0]
    result = []
    while ready:
        layer = ready.copy()
        result.append(layer)
        ready.clear()
        for dep in layer:
            for node in in_edges[dep]:
                out_degrees[node] -= 1
                if out_degrees[node] == 0:
                    ready.append(node)

    if len(workflow) != sum(map(len, result)):
        raise ValueError("cyclic dependency detected")

    hierarchy = [list(map(workflow.get, layer)) for layer in result]
    reachable = {item.key for item in workflow}
    return ToposortResult(in_edges, hierarchy, edge_labels, reachable)


def get_dependency_path(workflow: Flow, target: str) -> ToposortResult:
    reachable = workflow.get_reachable(target)
    in_edges, raw_hierarchy, edge_labels, _ = toposort(workflow)
    hierarchy = []
    for raw_layer in raw_hierarchy:
        layer = []
        for item in raw_layer:
            if item.key in reachable:
                layer.append(item)
        if layer:
            hierarchy.append(layer)
    return ToposortResult(in_edges, hierarchy, edge_labels, reachable)


def render_workflow(
    workflow: Flow,
    *,
    target: Optional[str] = None,
    fig_w_ratio: int = 4,
    fig_h_ratio: int = 3,
    dpi: int = 200,
    node_size: int = 2000,
    node_shape: str = "s",
    node_color: str = "lightblue",
    layout: str = "multipartite_layout",
) -> Image.Image:
    # setup graph
    G = nx.DiGraph()
    if target is None:
        target = workflow.last.key
    in_edges, hierarchy, edge_labels, _ = get_dependency_path(workflow, target)
    # setup plt
    fig_w = max(fig_w_ratio * len(hierarchy), 8)
    fig_h = fig_h_ratio * max(map(len, hierarchy))
    plt.figure(figsize=(fig_w, fig_h), dpi=dpi)
    box = plt.gca().get_position()
    plt.gca().set_position([box.x0, box.y0, box.width * 0.8, box.height])
    # map key to indices
    key2idx: Dict[str, int] = {}
    for layer in hierarchy:
        for node in layer:
            key2idx[node.key] = len(key2idx)
    # add nodes
    for i, layer in enumerate(hierarchy):
        for node in layer:
            G.add_node(key2idx[node.key], subset=f"layer_{i}")
    # add edges
    for dep, links in in_edges.items():
        for link in links:
            if dep not in key2idx or link not in key2idx:
                continue
            label = edge_labels[(link, dep)]
            G.add_edge(key2idx[dep], key2idx[link], label=label)
    # calculate positions
    layout_fn = getattr(nx, layout, None)
    if layout_fn is None:
        raise ValueError(f"unknown layout: {layout}")
    pos = layout_fn(G)
    # draw the nodes
    nodes_styles = dict(
        node_size=node_size,
        node_shape=node_shape,
        node_color=node_color,
    )
    nx.draw_networkx_nodes(G, pos, **nodes_styles)
    node_labels_styles = dict(
        font_size=18,
    )
    nx.draw_networkx_labels(G, pos, **node_labels_styles)
    # draw the edges
    nx_edge_labels = nx.get_edge_attributes(G, "label")
    nx.draw_networkx_edges(
        G,
        pos,
        arrows=True,
        arrowstyle="-|>",
        arrowsize=16,
        node_size=nodes_styles["node_size"],
        node_shape=nodes_styles["node_shape"],
    )
    nx.draw_networkx_edge_labels(G, pos, edge_labels=nx_edge_labels)
    # draw captions
    patches = [
        mpatches.Patch(color=node_color, label=f"{idx}: {key}")
        for key, idx in key2idx.items()
    ]
    plt.legend(handles=patches, bbox_to_anchor=(1, 0.5), loc="center left")
    # render
    plt.axis("off")
    buf = io.BytesIO()
    plt.savefig(buf, format="png")
    buf.seek(0)
    return Image.open(buf)


__all__ = [
    "toposort",
    "get_dependency_path",
    "render_workflow",
]



from setuptools import setup, find_packages

VERSION = "0.1.0"
PACKAGE_NAME = "carefree-workflow"

DESCRIPTION = "Build arbitray workflows with Python!"
with open("README.md", encoding="utf-8") as f:
    LONG_DESCRIPTION = f.read()

setup(
    name=PACKAGE_NAME,
    version=VERSION,
    packages=find_packages(exclude=("tests",)),
    include_package_data=True,
    entry_points={"console_scripts": ["cflow = cflow.cli:main"]},
    install_requires=[
        "pillow",
        "aiohttp",
        "fastapi",
        "uvicorn",
        "networkx",
        "requests",
        "matplotlib",
        "opencv-python-headless",
        "click>=8.1.3",
        "carefree-toolkit>=0.3.11",
    ],
    author="carefree0910",
    author_email="syameimaru.saki@gmail.com",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    keywords="python workflow",
)
